'''
线程类 MyThread，以及线程管理类MyProcess

MyThread 创建线程，并获取执行结果
MyProcess 从线程池创建/获取线程。阻塞运行，直到所有线程执行完毕，统一返回结果
'''

import threading
import time


class MyThread(threading.Thread):
    def __init__(self, func, args=(), semaphore=None):
        super(MyThread, self).__init__()
        self._func = func
        self._args = args
        self._result = None
        self._semaphore = semaphore

    def reset(self, func, args=()):
        self._started.clear()
        self._func = func
        self._args = args

    def start(self):
        if self._semaphore:
            self._semaphore.acquire()
        super(MyThread, self).start()

    def run(self):
        self._result = self._func(*self._args)
        if self._semaphore:
            self._semaphore.release()

    def get_result(self):
        try:
            return self._result
        except Exception:
            return None


class MyProcess(object):
    def __init__(self, batch=10, semaphore=None):
        self._threads_pool = {}  # 线程池，重复使用，可避免创建新线程导致的内存占用持续增加
        self._batch = batch
        self._start_index = 0
        self._end_index = batch
        self._semaphore = semaphore

    def __del__(self):
        self._semaphore = None
        self._threads_pool = None

    def begin(self, data, func, each_callback=None, finish_callback=None):
        '''
        :param data: 需要处理的数据，数组
        :param func: 处理数据的方法
        :param each_callback: 每处理一个数据，回调方法
        :param finish_callback: 处理完所有数据，回调方法
        :return: 所有数据处理的结果。
                 等同于`finish_callback`的回调参数
        '''
        results = []
        while True:
            if not data:
                break
            sub = data[:self._batch]
            data = data[self._batch:]
            threads = []
            for i in range(len(sub)):
                arg = sub[i]
                thread = self._threads_pool.get(i)
                if thread:
                    thread.reset(func, args=(arg,))
                else:
                    thread = MyThread(func, args=(arg,), semaphore=self._semaphore)
                    self._threads_pool[i] = thread
                thread.start()
                threads.append(thread)
            for thread in threads:
                thread.join()
            for thread in threads:
                result = thread.get_result()
                # result 为 None的情况，外面处理
                if callable(each_callback):
                    each_callback(result)  # 回调本次结果，外面根据自己情况处理
                # 这里只是 append
                # 当 result 为 list 时，
                # 可能需要的是 extend，还可以在外部自行处理
                # 但无法具体分辨实际需求
                results.append(result)
        if callable(finish_callback):
            finish_callback(results)
        return results  # 返回值可用，可不用

    def begin_v2(self, data, func, each_callback=None, finish_callback=None):
        ''' v2 - func处理函数的参数为 index, item
        :param data: 需要处理的数据，数组
        :param func: 处理数据的方法
        :param each_callback: 每处理一个数据，回调方法
        :param finish_callback: 处理完所有数据，回调方法
        :return: 所有数据处理的结果。
                 等同于`finish_callback`的回调参数
        '''
        results = []
        while True:
            if self._start_index >= len(data):
                break
            threads = []
            for i in range(self._start_index, self._end_index):
                arg = data[i]
                thread = self._threads_pool.get(i%self._batch)
                if thread:
                    thread.reset(func, args=(i, arg,))
                else:
                    thread = MyThread(func, args=(i, arg,), semaphore=self._semaphore)
                    self._threads_pool[i%self._batch] = thread
                thread.start()
                threads.append(thread)
            # 更新下标
            self._start_index = self._end_index
            self._end_index = self._end_index + self._batch
            for thread in threads:
                thread.join()
            for thread in threads:
                result = thread.get_result()
                # result 为 None的情况，外面处理
                if callable(each_callback):
                    each_callback(result)  # 回调本次结果，外面根据自己情况处理
                # 这里只是 append
                # 当 result 为 list 时，
                # 可能需要的是 extend，还可以在外部自行处理
                # 但无法具体分辨实际需求
                results.append(result)
        self._start_index = 0
        self._end_index = self._batch
        if callable(finish_callback):
            finish_callback(results)
        return results  # 返回值可用，可不用

###
### test
###

def go_v2(index, param):
    # 单个线程需要进行的操作
    ret = '%04d-%04d'%(index, param)
    time.sleep(1)
    return ret

def go(param):
    ret = '%04d'%param
    time.sleep(1)
    return ret

def each(result):
    print(result)

def finish(results):
    print('finish', len(results))


if __name__ == '__main__':
    # semaphore = threading.Semaphore(15)  # 限制，也是保护
    process = MyProcess(batch=5)

    while True:
        arr = []
        for i in range(20):
            arr.append(i)

        ret = process.begin_v2(arr, go_v2, each, finish)
        time.sleep(3)
    pass



import rsa
import base64


# encrypt

class Encryptor(object):
    def __init__(self):
        self._private_key = None
        self._public_key = None

    def set_key(self, public, private=None):
        if not public:
            raise Exception('public key is None')
        is_rsa = 'BEGIN RSA PUBLIC KEY' in public  # 关键是有没有`RSA`
        if is_rsa:
            self._public_key = rsa.PublicKey.load_pkcs1(public.encode())
            self._private_key = private and rsa.PrivateKey.load_pkcs1(private.encode())
        else:
            self._public_key = rsa.PublicKey.load_pkcs1_openssl_pem(public.encode())
            self._private_key = private and rsa.PrivateKey.load_pkcs1_openssl_pem(private.encode())

    def set_pem(self, public, private=None):
        # public
        try:
            with open(public, 'r') as file:
                public_key_str = file.read()
        except:
            public_key_str = None
        # private
        try:
            with open(private, 'r') as file:
                private_key_str = file.read()
        except:
            private_key_str = None
        # set key
        self.set_key(public_key_str, private_key_str)

    def encrypt(self, content: str) -> str:
        encrypted = rsa.encrypt(content.encode(), self._public_key)
        result = base64.b64encode(encrypted).decode('utf-8')
        return result

    def decrypt(self, secret: str) -> str:
        decrypted = base64.b64decode(secret.encode())
        result = rsa.decrypt(decrypted, self._private_key).decode()
        return result


def generate_pem_key(name: str='') -> (rsa.PublicKey, rsa.PrivateKey):
    public_key, private_key = rsa.newkeys(1024)
    public = public_key.save_pkcs1('PEM')  # default 'PEM'
    private = private_key.save_pkcs1('PEM')
    if name:
        prefix = name + '_'
        with open(prefix+'public_key.pem', 'wb') as file:
            file.write(public)
        with open(prefix+'private_key.pem', 'wb') as file:
            file.write(private)
    return public_key, private_key


# test


def test1():
    # public, private = generate_pem_key('test')
    public = '''
    -----BEGIN RSA PUBLIC KEY-----
    MIGJAoGBAIDZn/enhx0RRdNioXblQBywJJSuvcIwF8NPXcBaaplD6einSeC2NzoX
    N2sda4nO25hajiQBD41t2a0g5jP4nF1WWnya9ketnJXLZjotRd8VvxI3Swk+9tSQ
    vcJGBEfzhJ9iUuuygHM7SnxzPosDWm9VtYEJsUIavldLWZ4p4X/nAgMBAAE=
    -----END RSA PUBLIC KEY-----
    '''
    private = '''
    -----BEGIN RSA PRIVATE KEY-----
    MIICYAIBAAKBgQCA2Z/3p4cdEUXTYqF25UAcsCSUrr3CMBfDT13AWmqZQ+nop0ng
    tjc6FzdrHWuJztuYWo4kAQ+NbdmtIOYz+JxdVlp8mvZHrZyVy2Y6LUXfFb8SN0sJ
    PvbUkL3CRgRH84SfYlLrsoBzO0p8cz6LA1pvVbWBCbFCGr5XS1meKeF/5wIDAQAB
    AoGAG2CiObfR4J477OdHEYEydyYCD8l1Ll6Tnf8uF2HexoQEnld1PhbZczFdqBfP
    Mq/OPvf2vbWv/Uf6+WtE+zCWH9n1EG3zXme/DhQ2XYG7+unst4il+srtwrPkW8HA
    eci3Cgys0ximi7DkewLjKuJdS9/PcQf5WFZLllpythB5MgkCRQCmQVnvNuZPUl0Y
    WSDn5+9JYzC9pjS0o4ZS3MKSKET+hIellTRuAPgaK5os3a9DBwabSGo5ZN46J+Kt
    bSmxgxk0EN1VRQI9AMZnShZsGt3OK4vvj6Vmd+zfsdDhlbuGL2Ih/O91pWWh8J2B
    ANFw6vlqfXoDqHTegd8MG8IxFCPyP2qFOwJFAIL1QUk73minASvXsSLbQFJ3boJE
    tImBsaH9wMnuLIKrlEnq4JSx8Lx0kgo7SP2sQBj7DqlM+funRVfEgC4SjTzE+ANd
    Ajw+h6k6/eFNzL++v8bnGy9q0Wmqap6VVooyhIHCOrLhIDPEgDbwy4TTDPP085gx
    FTubP6a0AmHVnnDMMqcCRDv3fzyM1cFm3Uc2RV7IqJUC2ekVbDn2MKfUnBlJs6hk
    25fcLCtYJYYAZvLBcUg/bb7bNwZfAmTBWtYOxiXlO9JT9HAI
    -----END RSA PRIVATE KEY-----
    '''

    s = "hello, world"
    en = Encryptor()
    en.set_key(public, private)

    ss = en.encrypt(s)
    print(ss)

    s = en.decrypt(ss)
    print(s)


def test2():
    # public, private = generate_pem_key()
    s = "hello, world"
    en = Encryptor()
    en.set_pem('test_public_key.pem', 'test_private_key.pem')
    ss = en.encrypt(s)
    print(ss)

    s = en.decrypt(ss)
    print(s)


if __name__ == '__main__':
    test1()



from functools import wraps
from inspect import getattr_static


'''
为自定义类注册方法的修饰器

声明一个空类，没有定义任何方法
    class myclass(object):
        pass

注册一个方法
    @method_register(myclass)
    def hi(self):
        print('hello')

调用新方法
    obj = myclass()
    obj.hi()

'''


def method_register(cls):
    def decorator(func):
        @wraps(func) 
        def wrapper(self, *args, **kwargs): 
            return func(self, *args, **kwargs)
        if getattr_static(cls, func.__name__, None):
            msg = 'Error method name REPEAT, {} has exist'.format(func.__name__)
            raise NameError(msg)
        else:
            setattr(cls, func.__name__, wrapper)
        return func 
    return decorator



import logging
import datetime
import os


class Logger(object):
    __msgfmt = '%(asctime)s.%(msecs)03d | %(filename)s:%(funcName)s[%(lineno)d] | %(levelname)s : %(message)s'
    __datefmt = '%Y-%m-%d %H:%M:%S'

    def __init__(self, filename=None):
        if filename is None:
            filename = ''
        logname = '{:%Y%m%d%H%M%S%f}'.format(datetime.datetime.now())+filename
        self.logger = logging.getLogger(logname)
        self.__checkFilePath(filename)
        self.__setupFileLog(filename)
        self.__setupConsoleLog()

    def __setupFileLog(self, file):
        if file is None or len(file) == 0:
            return
        formatter = logging.Formatter(self.__msgfmt)
        formatter.datefmt = self.__datefmt
        file_handler = logging.FileHandler(file)
        file_handler.setFormatter(formatter)
        file_handler.setLevel(logging.INFO)
        self.logger.addHandler(file_handler)

    def __setupConsoleLog(self):
        formatter = logging.Formatter(self.__msgfmt)
        formatter.datefmt = self.__datefmt
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(console_handler)

    def __checkFilePath(self, filename):
        arr = filename.split('/')
        if len(arr) == 1:
            return
        makeFilePath(filename)

    def d(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.debug(msg, stacklevel=2)

    def i(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.info(msg, stacklevel=2)

    def w(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.warning(msg, stacklevel=2)

    def e(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.error(msg, stacklevel=2)

    def exc(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.exception(msg, stacklevel=3)

    def x(self, *args):
        msg = ' '.join(['%s'%i for i in args])
        self.logger.critical(msg, stacklevel=2)


def makeFilePath(path):
    arr = path.split('/')
    if len(arr) == 1:
        return
    filepath = os.path.join(*arr[:-1])
    if os.path.exists(filepath):
        return
    paths = filepath.split('/')
    for i in range(len(paths)):
        folder = '/'.join(paths[:i + 1])
        if not os.path.exists(folder):
            os.mkdir(folder)
    return filepath


def test():
    log = Logger()
    log.d('debug', 1, 'a', {'k':'v'})
    log.i('info', 'some info')
    log.e('error', 123, 3*5)
    try:
        a = 1/0
        log.d(a)
    except:
        log.exc('做除法的时候出错')
    pass


if __name__ == '__main__':
    test()
